



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %SP_DIR%\\rfi_file_monitor\\docs\\queue_manager.pango exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %SP_DIR%\\rfi_file_monitor\\operations\\docs\\\template.pango exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
if not exist %SP_DIR%\\rfi_file_monitor\\engines\\docs\\\template.pango exit 1
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
